"""A simple transformation that truncates the text of input documents to a
given character length."""

from annif.exception import ConfigurationException
from . import transform


class InputLimiter(transform.BaseTransform):

    name = 'limit'

    def __init__(self, project, input_limit):
        super().__init__(project)
        self.input_limit = int(input_limit)
        self._validate_value(self.input_limit)

    def transform_fn(self, text):
        return text[:self.input_limit]

    def _validate_value(self, input_limit):
        if input_limit < 0:
            raise ConfigurationException(
                'input_limit in limit_input transform cannot be negative',
                project_id=self.project.project_id)



"""Parallel processing functionality for Annif"""


import multiprocessing
import multiprocessing.dummy


class BaseWorker:
    """Base class for workers that implement tasks executed via
    multiprocessing. The init method can be used to store data objects that
    are necessary for the operation. They will be stored in a class
    attribute that is accessible to the static worker method. The storage
    solution is inspired by this blog post:
    https://thelaziestprogrammer.com/python/multiprocessing-pool-a-global-solution # noqa
    """

    args = None

    @classmethod
    def init(cls, args):
        cls.args = args


class ProjectSuggestMap:
    """A utility class that can be used to wrap one or more projects and
    provide a mapping method that converts Document objects to suggestions.
    Intended to be used with the multiprocessing module."""

    def __init__(
            self,
            registry,
            project_ids,
            backend_params,
            limit,
            threshold):
        self.registry = registry
        self.project_ids = project_ids
        self.backend_params = backend_params
        self.limit = limit
        self.threshold = threshold

    def suggest(self, doc):
        filtered_hits = {}
        for project_id in self.project_ids:
            project = self.registry.get_project(project_id)
            hits = project.suggest(doc.text, self.backend_params)
            filtered_hits[project_id] = hits.filter(
                project.subjects, self.limit, self.threshold)
        return (filtered_hits, doc.uris, doc.labels)


def get_pool(n_jobs):
    """return a suitable multiprocessing pool class, and the correct jobs
    argument for its constructor, for the given amount of parallel jobs"""

    if n_jobs < 1:
        n_jobs = None
        pool_class = multiprocessing.Pool
    elif n_jobs == 1:
        # use the dummy wrapper around threading to avoid subprocess overhead
        pool_class = multiprocessing.dummy.Pool
    else:
        pool_class = multiprocessing.Pool

    return n_jobs, pool_class



"""Unit tests for the language-filter transform in Annif"""

import pytest
import annif.transform

pytest.importorskip("annif.transform.langfilter")


def test_lang_filter(project):
    transf = annif.transform.get_transform("filter_lang", project)
    text = """
        Kansalliskirjasto on kaikille avoin kulttuuriperintöorganisaatio, joka
        palvelee valtakunnallisesti kansalaisia, tiedeyhteisöjä ja muita
        yhteiskunnan toimijoita.
        The National Library of Finland is the oldest and largest scholarly
        library in Finland. It is responsible for the collection, description,
        preservation and accessibility of Finland’s published national heritage
        and the unique collections under its care.
        Nationalbiblioteket är Finlands största och äldsta vetenskapliga
        bibliotek, som ansvarar för utökning, beskrivning, förvaring och
        tillhandahållande av vårt nationella publikationsarv och av sina unika
        samlingar.
        Abc defghij klmnopqr stuwxyz abc defghij klmnopqr stuwxyz.
        Turvaamme Suomessa julkaistun tai Suomea koskevan julkaistun
        kulttuuriperinnön saatavuuden sekä välittämme ja tuotamme
        tietosisältöjä tutkimukselle, opiskelulle, kansalaisille ja
        yhteiskunnalle. Kehitämme palveluja yhteistyössä kirjastojen,
        arkistojen, museoiden ja muiden toimijoiden kanssa.
    """
    text = ' '.join(text.split())
    text_filtered = """
        Kansalliskirjasto on kaikille avoin kulttuuriperintöorganisaatio, joka
        palvelee valtakunnallisesti kansalaisia, tiedeyhteisöjä ja muita
        yhteiskunnan toimijoita.
        Abc defghij klmnopqr stuwxyz abc defghij klmnopqr stuwxyz.
        Turvaamme Suomessa julkaistun tai Suomea koskevan julkaistun
        kulttuuriperinnön saatavuuden sekä välittämme ja tuotamme
        tietosisältöjä tutkimukselle, opiskelulle, kansalaisille ja
        yhteiskunnalle. Kehitämme palveluja yhteistyössä kirjastojen,
        arkistojen, museoiden ja muiden toimijoiden kanssa.
    """
    text_filtered = ' '.join(text_filtered.split())
    assert transf.transform_text(text) == text_filtered


def test_lang_filter_text_min_length(project):
    text = "This is just some non-Finnish text of 52 characters."
    transf = annif.transform.get_transform("filter_lang", project)
    assert transf.transform_text(text) == text
    # Set a short text_min_length to apply language filtering:
    transf = annif.transform.get_transform(
        "filter_lang(text_min_length=50)", project)
    assert transf.transform_text(text) == ""


def test_lang_filter_sentence_min_length(project):
    text = "This is a non-Finnish sentence of 42 chars. And this of 20 chars."
    transf = annif.transform.get_transform(
        "filter_lang(text_min_length=50)", project)
    assert transf.transform_text(text) == text
    # Set a short sentence_min_length to apply language filtering:
    transf = annif.transform.get_transform(
        "filter_lang(text_min_length=50,sentence_min_length=30)", project)
    assert transf.transform_text(text) == "And this of 20 chars."



"""MLLM (Maui-like Lexical Matchin) model for Annif"""

import collections
import math
import joblib
from statistics import mean
from enum import IntEnum
import numpy as np
from rdflib.namespace import SKOS
from sklearn.feature_extraction.text import CountVectorizer
from sklearn.ensemble import BaggingClassifier
from sklearn.tree import DecisionTreeClassifier
import annif.util
import annif.parallel
from annif.exception import OperationFailedException
from annif.lexical.tokenset import TokenSet, TokenSetIndex
from annif.lexical.util import get_subject_labels
from annif.lexical.util import make_relation_matrix, make_collection_matrix


Term = collections.namedtuple('Term', 'subject_id label is_pref')

Match = collections.namedtuple(
    'Match', 'subject_id is_pref n_tokens pos ambiguity')

Candidate = collections.namedtuple(
    'Candidate',
    'doc_length subject_id freq is_pref n_tokens ambiguity ' +
    'first_occ last_occ spread')

ModelData = collections.namedtuple(
    'ModelData',
    'broader narrower related collection ' +
    'doc_freq subj_freq idf')

Feature = IntEnum(
    'Feature',
    'freq doc_freq subj_freq tfidf is_pref n_tokens ambiguity ' +
    'first_occ last_occ spread doc_length ' +
    'broader narrower related collection',
    start=0)


def conflate_matches(matches, doc_length):
    subj_matches = collections.defaultdict(list)
    for match in matches:
        subj_matches[match.subject_id].append(match)
    return [
        Candidate(
            doc_length=doc_length,
            subject_id=subject_id,
            freq=len(matches) / doc_length,
            is_pref=mean((float(m.is_pref) for m in matches)),
            n_tokens=mean((m.n_tokens for m in matches)),
            ambiguity=mean((m.ambiguity for m in matches)),
            first_occ=matches[0].pos / doc_length,
            last_occ=matches[-1].pos / doc_length,
            spread=(matches[-1].pos - matches[0].pos) / doc_length
        )
        for subject_id, matches in subj_matches.items()]


def generate_candidates(text, analyzer, vectorizer, index):
    sentences = analyzer.tokenize_sentences(text)
    sent_tokens = vectorizer.transform(sentences)
    matches = []

    for sent_idx, token_matrix in enumerate(sent_tokens):
        tset = TokenSet(token_matrix.nonzero()[1])
        for ts, ambiguity in index.search(tset):
            matches.append(Match(subject_id=ts.subject_id,
                                 is_pref=ts.is_pref,
                                 n_tokens=len(ts),
                                 pos=sent_idx,
                                 ambiguity=ambiguity))

    return conflate_matches(matches, len(sentences))


def candidates_to_features(candidates, mdata):
    """Convert a list of Candidates to a NumPy feature matrix"""

    matrix = np.zeros((len(candidates), len(Feature)), dtype=np.float32)
    c_ids = [c.subject_id for c in candidates]
    c_vec = np.zeros(mdata.related.shape[0], dtype=np.bool)
    c_vec[c_ids] = True
    broader = mdata.broader.multiply(c_vec).sum(axis=1)
    narrower = mdata.narrower.multiply(c_vec).sum(axis=1)
    related = mdata.related.multiply(c_vec).sum(axis=1)
    collection = mdata.collection.multiply(c_vec).T.dot(
        mdata.collection).sum(axis=0)
    for idx, c in enumerate(candidates):
        subj = c.subject_id
        matrix[idx, Feature.freq] = c.freq
        matrix[idx, Feature.doc_freq] = mdata.doc_freq[subj]
        matrix[idx, Feature.subj_freq] = mdata.subj_freq.get(subj, 1) - 1
        matrix[idx, Feature.tfidf] = c.freq * mdata.idf[subj]
        matrix[idx, Feature.is_pref] = c.is_pref
        matrix[idx, Feature.n_tokens] = c.n_tokens
        matrix[idx, Feature.ambiguity] = c.ambiguity
        matrix[idx, Feature.first_occ] = c.first_occ
        matrix[idx, Feature.last_occ] = c.last_occ
        matrix[idx, Feature.spread] = c.spread
        matrix[idx, Feature.doc_length] = c.doc_length
        matrix[idx, Feature.broader] = broader[subj, 0] / len(c_ids)
        matrix[idx, Feature.narrower] = narrower[subj, 0] / len(c_ids)
        matrix[idx, Feature.related] = related[subj, 0] / len(c_ids)
        matrix[idx, Feature.collection] = collection[0, subj] / len(c_ids)
    return matrix


class MLLMCandidateGenerator(annif.parallel.BaseWorker):

    @classmethod
    def generate_candidates(cls, doc_subject_ids, text):
        candidates = generate_candidates(text, **cls.args)
        return doc_subject_ids, candidates


class MLLMFeatureConverter(annif.parallel.BaseWorker):

    @classmethod
    def candidates_to_features(cls, candidates):
        return candidates_to_features(candidates,
                                      **cls.args)  # pragma: no cover


class MLLMModel:
    """Maui-like Lexical Matching model"""

    def generate_candidates(self, text, analyzer):
        return generate_candidates(text, analyzer,
                                   self._vectorizer, self._index)

    @property
    def _model_data(self):
        return ModelData(broader=self._broader_matrix,
                         narrower=self._narrower_matrix,
                         related=self._related_matrix,
                         collection=self._collection_matrix,
                         doc_freq=self._doc_freq,
                         subj_freq=self._subj_freq,
                         idf=self._idf)

    def _candidates_to_features(self, candidates):
        return candidates_to_features(candidates, self._model_data)

    def _prepare_terms(self, graph, vocab, params):
        if annif.util.boolean(params['use_hidden_labels']):
            label_props = [SKOS.altLabel, SKOS.hiddenLabel]
        else:
            label_props = [SKOS.altLabel]

        terms = []
        subject_ids = []
        for subj_id, uri, pref, _ in vocab.subjects.active:
            subject_ids.append(subj_id)
            terms.append(Term(subject_id=subj_id, label=pref, is_pref=True))

            for label in get_subject_labels(graph, uri, label_props,
                                            params['language']):
                terms.append(Term(subject_id=subj_id,
                                  label=label,
                                  is_pref=False))

        return (terms, subject_ids)

    def _prepare_relations(self, graph, vocab):
        self._broader_matrix = make_relation_matrix(
            graph, vocab, SKOS.broader)
        self._narrower_matrix = make_relation_matrix(
            graph, vocab, SKOS.narrower)
        self._related_matrix = make_relation_matrix(
            graph, vocab, SKOS.related)
        self._collection_matrix = make_collection_matrix(graph, vocab)

    def _prepare_train_index(self, vocab, analyzer, params):
        graph = vocab.as_graph()
        terms, subject_ids = self._prepare_terms(graph, vocab, params)
        self._prepare_relations(graph, vocab)

        self._vectorizer = CountVectorizer(
            binary=True,
            tokenizer=analyzer.tokenize_words
        )
        label_corpus = self._vectorizer.fit_transform((t.label for t in terms))

        # frequency of each token used in labels - how rare each word is
        token_freq = np.bincount(label_corpus.indices,
                                 minlength=label_corpus.shape[1])

        self._index = TokenSetIndex()
        for term, label_matrix in zip(terms, label_corpus):
            tokens = label_matrix.nonzero()[1]
            # sort tokens by frequency - use the rarest token as index key
            tokens = sorted(tokens, key=token_freq.__getitem__)
            tset = TokenSet(tokens, term.subject_id, term.is_pref)
            self._index.add(tset)

        return subject_ids

    def _prepare_train_data(self, corpus, vocab, analyzer, n_jobs):
        # frequency of subjects (by id) in the generated candidates
        self._doc_freq = collections.Counter()
        # frequency of manually assigned subjects ("domain keyphraseness")
        self._subj_freq = collections.Counter()
        train_x = []
        train_y = []

        jobs, pool_class = annif.parallel.get_pool(n_jobs)

        cg_args = {
            'analyzer': analyzer,
            'vectorizer': self._vectorizer,
            'index': self._index
        }

        with pool_class(jobs,
                        initializer=MLLMCandidateGenerator.init,
                        initargs=(cg_args,)) as pool:
            params = (([vocab.subjects.by_uri(uri) for uri in doc.uris],
                       doc.text)
                      for doc in corpus.documents)
            for doc_subject_ids, candidates in pool.starmap(
                    MLLMCandidateGenerator.generate_candidates, params, 10):

                self._subj_freq.update(doc_subject_ids)
                self._doc_freq.update([c.subject_id for c in candidates])
                train_x.append(candidates)
                train_y += [(c.subject_id in doc_subject_ids)
                            for c in candidates]

        return (train_x, train_y)

    def _calculate_idf(self, subject_ids, doc_count):
        idf = collections.defaultdict(float)
        for subj_id in subject_ids:
            idf[subj_id] = math.log((doc_count + 1) /
                                    (self._doc_freq[subj_id] + 1)) + 1

        return idf

    def _prepare_features(self, train_x, n_jobs):
        fc_args = {'mdata': self._model_data}
        jobs, pool_class = annif.parallel.get_pool(n_jobs)

        with pool_class(jobs,
                        initializer=MLLMFeatureConverter.init,
                        initargs=(fc_args,)) as pool:
            features = pool.map(
                MLLMFeatureConverter.candidates_to_features, train_x, 10)

        return features

    def prepare_train(self, corpus, vocab, analyzer, params, n_jobs):
        # create an index from the vocabulary terms
        subject_ids = self._prepare_train_index(vocab, analyzer, params)

        # convert the corpus into train data
        train_x, train_y = self._prepare_train_data(
            corpus, vocab, analyzer, n_jobs)

        # precalculate idf values for all candidate subjects
        self._idf = self._calculate_idf(subject_ids, len(train_x))

        # convert the train data into feature values
        features = self._prepare_features(train_x, n_jobs)

        return (np.vstack(features), np.array(train_y))

    def _create_classifier(self, params):
        return BaggingClassifier(
            DecisionTreeClassifier(
                min_samples_leaf=int(params['min_samples_leaf']),
                max_leaf_nodes=int(params['max_leaf_nodes'])
            ), max_samples=float(params['max_samples']))

    def train(self, train_x, train_y, params):
        # fit the model on the training corpus
        self._classifier = self._create_classifier(params)
        self._classifier.fit(train_x, train_y)
        # sanity check: verify that the classifier has seen both classes
        if self._classifier.n_classes_ != 2:
            raise OperationFailedException(
                "Unable to create classifier: " +
                "Not enough positive and negative examples " +
                "in the training data. Please check that your training " +
                "data matches your vocabulary.")

    def _prediction_to_list(self, scores, candidates):
        subj_scores = [(score[1], c.subject_id)
                       for score, c in zip(scores, candidates)]
        return sorted(subj_scores, reverse=True)

    def predict(self, candidates):
        if not candidates:
            return []
        features = self._candidates_to_features(candidates)
        scores = self._classifier.predict_proba(features)
        return self._prediction_to_list(scores, candidates)

    def save(self, filename):
        return joblib.dump(self, filename)

    @staticmethod
    def load(filename):
        return joblib.load(filename)



"""Functionality for obtaining text transformation from string specification"""

import re
import annif
from . import transform
from . import inputlimiter
from annif.util import parse_args
from annif.exception import ConfigurationException


def parse_specs(transform_specs):
    """Parse a transformation specification into a list of tuples, e.g.
    'transf_1(x),transf_2(y=42),transf_3' is parsed to
    [(transf_1, [x], {}), (transf_2, [], {y: 42}), (transf_3, [], {})]."""

    parsed = []
    # Split by commas not inside parentheses
    parts = re.split(r',\s*(?![^()]*\))', transform_specs)
    for part in parts:
        match = re.match(r'(\w+)(\((.*)\))?', part)
        if match is None:
            continue
        transform = match.group(1)
        posargs, kwargs = parse_args(match.group(3))
        parsed.append((transform, posargs, kwargs))
    return parsed


def get_transform(transform_specs, project):
    transform_defs = parse_specs(transform_specs)
    transform_classes = []
    args = []
    for trans, posargs, kwargs in transform_defs:
        if trans not in _transforms:
            raise ConfigurationException(f"No such transform {trans}")
        transform_classes.append(_transforms[trans])
        args.append((posargs, kwargs))
    return transform.TransformChain(transform_classes, args, project)


_transforms = {
    transform.IdentityTransform.name: transform.IdentityTransform,
    inputlimiter.InputLimiter.name: inputlimiter.InputLimiter}

# Optional transforms
try:
    from . import langfilter
    _transforms.update({langfilter.LangFilter.name: langfilter.LangFilter})
except ImportError:
    annif.logger.debug(
        "pycld3 not available, not enabling filter_language transform")



"""Unit tests for the SVC backend in Annif"""

import pytest
import annif.backend
import annif.corpus
from annif.exception import NotInitializedException
from annif.exception import NotSupportedException


def test_svc_default_params(project):
    svc_type = annif.backend.get_backend("svc")
    svc = svc_type(
        backend_id='svc',
        config_params={},
        project=project)

    expected_default_params = {
        'limit': 100,
        'min_df': 1,
    }
    actual_params = svc.params
    for param, val in expected_default_params.items():
        assert param in actual_params and actual_params[param] == val


def test_svc_suggest_no_vectorizer(project):
    svc_type = annif.backend.get_backend('svc')
    svc = svc_type(
        backend_id='svc',
        config_params={},
        project=project)

    with pytest.raises(NotInitializedException):
        svc.suggest("example text")


def test_svc_train(datadir, document_corpus, project, caplog):
    svc_type = annif.backend.get_backend('svc')
    svc = svc_type(
        backend_id='svc',
        config_params={},
        project=project)

    svc.train(document_corpus)
    assert svc._model is not None
    assert datadir.join('svc-model.gz').exists()
    assert 'training on a document with multiple subjects is not ' + \
           'supported by SVC; selecting one random subject.' in caplog.text


def test_svc_train_ngram(datadir, document_corpus, project):
    svc_type = annif.backend.get_backend('svc')
    svc = svc_type(
        backend_id='svc',
        config_params={'ngram': 2},
        project=project)

    svc.train(document_corpus)
    assert svc._model is not None
    assert datadir.join('svc-model.gz').exists()


def test_svc_train_cached(datadir, project):
    svc_type = annif.backend.get_backend('svc')
    svc = svc_type(
        backend_id='svc',
        config_params={},
        project=project)

    with pytest.raises(NotSupportedException):
        svc.train("cached")


def test_svc_train_nodocuments(datadir, project, empty_corpus):
    svc_type = annif.backend.get_backend('svc')
    svc = svc_type(
        backend_id='svc',
        config_params={},
        project=project)

    with pytest.raises(NotSupportedException):
        svc.train(empty_corpus)


def test_svc_suggest(project):
    svc_type = annif.backend.get_backend('svc')
    svc = svc_type(
        backend_id='svc',
        config_params={'limit': 20},
        project=project)

    results = svc.suggest("""Arkeologiaa sanotaan joskus myös...""")

    assert len(results) > 0
    assert len(results) <= 20
    hits = results.as_list(project.subjects)
    assert 'http://www.yso.fi/onto/yso/p10849' in [
        result.uri for result in hits]
    assert 'arkeologit' in [result.label for result in hits]


def test_svc_suggest_no_input(project):
    svc_type = annif.backend.get_backend('svc')
    svc = svc_type(
        backend_id='svc',
        config_params={'limit': 8},
        project=project)

    results = svc.suggest("j")
    assert len(results) == 0


def test_svc_suggest_no_model(datadir, project):
    svc_type = annif.backend.get_backend('svc')
    svc = svc_type(
        backend_id='svc',
        config_params={},
        project=project)

    datadir.join('svc-model.gz').remove()
    with pytest.raises(NotInitializedException):
        svc.suggest("example text")



"""Common functionality for backends."""

import abc
import os.path
from datetime import datetime, timezone
from glob import glob
from annif import logger


class AnnifBackend(metaclass=abc.ABCMeta):
    """Base class for Annif backends that perform analysis. The
    non-implemented methods should be overridden in subclasses."""

    name = None
    needs_subject_index = False

    DEFAULT_PARAMETERS = {'limit': 100}

    def __init__(self, backend_id, config_params, project):
        """Initialize backend with specific parameters. The
        parameters are a dict. Keys and values depend on the specific
        backend type."""
        self.backend_id = backend_id
        self.config_params = config_params
        self.project = project
        self.datadir = project.datadir

    def default_params(self):
        return self.DEFAULT_PARAMETERS

    @property
    def params(self):
        params = {}
        params.update(self.default_params())
        params.update(self.config_params)
        return params

    @property
    def is_trained(self):
        return bool(glob(os.path.join(self.datadir, '*')))

    @property
    def modification_time(self):
        mtimes = [datetime.utcfromtimestamp(os.path.getmtime(p))
                  for p in glob(os.path.join(self.datadir, '*'))]
        most_recent = max(mtimes, default=None)
        if most_recent is None:
            return None
        return most_recent.replace(tzinfo=timezone.utc)

    def _get_backend_params(self, params):
        backend_params = dict(self.params)
        if params is not None:
            backend_params.update(params)
        return backend_params

    def _train(self, corpus, params, jobs=0):
        """This method can be overridden by backends. It implements
        the train functionality, with pre-processed parameters."""
        pass  # default is to do nothing, subclasses may override

    def train(self, corpus, params=None, jobs=0):
        """Train the model on the given document or subject corpus."""
        beparams = self._get_backend_params(params)
        return self._train(corpus, params=beparams, jobs=jobs)

    def initialize(self, parallel=False):
        """This method can be overridden by backends. It should cause the
        backend to pre-load all data it needs during operation.
        If parallel is True, the backend should expect to be used for
        parallel operation."""
        pass

    @abc.abstractmethod
    def _suggest(self, text, params):
        """This method should implemented by backends. It implements
        the suggest functionality, with pre-processed parameters."""
        pass  # pragma: no cover

    def suggest(self, text, params=None):
        """Suggest subjects for the input text and return a list of subjects
        represented as a list of SubjectSuggestion objects."""
        beparams = self._get_backend_params(params)
        self.initialize()
        return self._suggest(text, params=beparams)

    def debug(self, message):
        """Log a debug message from this backend"""
        logger.debug("Backend {}: {}".format(self.backend_id, message))

    def info(self, message):
        """Log an info message from this backend"""
        logger.info("Backend {}: {}".format(self.backend_id, message))

    def warning(self, message):
        """Log a warning message from this backend"""
        logger.warning("Backend {}: {}".format(self.backend_id, message))


class AnnifLearningBackend(AnnifBackend):
    """Base class for Annif backends that can perform online learning"""

    @abc.abstractmethod
    def _learn(self, corpus, params):
        """This method should implemented by backends. It implements the learn
        functionality, with pre-processed parameters."""
        pass  # pragma: no cover

    def learn(self, corpus, params=None):
        """Further train the model on the given document or subject corpus."""
        beparams = self._get_backend_params(params)
        return self._learn(corpus, params=beparams)



"""Unit tests for the input-transforms in Annif"""

import pytest
import annif.transform
from annif.exception import ConfigurationException
from annif.transform import parse_specs


def test_parse_specs():
    parsed = parse_specs('foo, bar(42,43,key=abc)')
    assert parsed == [('foo', [], {}), ('bar', ['42', '43'], {'key': 'abc'})]


def test_get_transform_nonexistent():
    with pytest.raises(ConfigurationException):
        annif.transform.get_transform("nonexistent", project=None)


def test_get_transform_badspec(project):
    with pytest.raises(ConfigurationException):
        annif.transform.get_transform("pass(invalid_argument)", project)


def test_input_limiter():
    transf = annif.transform.get_transform("limit(3)", project=None)
    assert transf.transform_text("running") == "run"


def test_input_limiter_with_negative_value(project):
    with pytest.raises(ConfigurationException):
        annif.transform.get_transform("limit(-2)", project)


def test_chained_transforms_text():
    transf = annif.transform.get_transform(
        "limit(5),pass,limit(3),", project=None)
    assert transf.transform_text("abcdefghij") == "abc"

    # Check with a more arbitrary transform function
    reverser = annif.transform.transform.IdentityTransform(None)
    reverser.transform_fn = lambda x: x[::-1]
    transf.transforms.append(reverser)
    assert transf.transform_text("abcdefghij") == "cba"


def test_chained_transforms_corpus(document_corpus):
    transf = annif.transform.get_transform(
        "limit(5),pass,limit(3),", project=None)
    transformed_corpus = transf.transform_corpus(document_corpus)
    for transf_doc, doc in zip(transformed_corpus.documents,
                               document_corpus.documents):
        assert transf_doc.text == doc.text[:3]
        assert transf_doc.uris == doc.uris
        assert transf_doc.labels == doc.labels

    # Check with a more arbitrary transform function
    reverser = annif.transform.transform.IdentityTransform(None)
    reverser.transform_fn = lambda x: x[::-1]
    transf.transforms.append(reverser)
    for transf_doc, doc in zip(transformed_corpus.documents,
                               document_corpus.documents):
        assert transf_doc.text == doc.text[:3][::-1]
        assert transf_doc.uris == doc.uris
        assert transf_doc.labels == doc.labels



"""Transformation filtering out parts of a text that are in a language
different from the language of the project."""

import annif
import cld3
from . import transform

logger = annif.logger


class LangFilter(transform.BaseTransform):

    name = 'filter_lang'

    def __init__(self, project, text_min_length=500, sentence_min_length=50):
        super().__init__(project)
        self.text_min_length = int(text_min_length)
        self.sentence_min_length = int(sentence_min_length)

    def _detect_language(self, text):
        """Tries to detect the language of a text input. Outputs a BCP-47-style
        language code (e.g. 'en')."""

        lan_info = cld3.get_language(text)
        if lan_info is not None and lan_info.is_reliable:
            return lan_info.language
        else:
            return None

    def transform_fn(self, text):
        if len(text) < self.text_min_length:
            return text

        retained_sentences = []
        for sent in self.project.analyzer.tokenize_sentences(text):
            if len(sent) < self.sentence_min_length:
                retained_sentences.append(sent)
                continue
            detected_lang = self._detect_language(sent)
            if detected_lang == self.project.language or detected_lang is None:
                retained_sentences.append(sent)
        return ' '.join(retained_sentences)
